#! -*- Encoding: Latin-1 -*-

def apriori_algorithmus(D, min_support = 0.50, min_confidence = 0.75):

    # ~~~~~~ Hilfsfunktion: ist A eine Teilmenge von B?  
    def is_subset(A, B):
        for item in A:
            if item not in B:
                return False
        return True

    # ~~~~~~ Hilfsfunktion: ist A gleich B?  
    def is_sameset(A, B):
        return is_subset(A, B) and is_subset(B, A)

    # ~~~~~~ Hilfsfunktion: erzeuge eine Menge A|B
    def mergesets(A, B):
        result = []
        for s in (A, B):
            for i in s:
                if i not in result:
                    result.append(i)
        return tuple(result)

    # ~~~~~~ Hilfsfunktion: Berechne den Support fr eine Menge s 
    def support(s):
        result = 0
        for item in D:
            if is_subset(s, item):
                result += 1
        return result/float(len(D))
        
    def confidence(r):
        n, m = 0, 0
        for item in D:
            if r[0] in item:
                n += 1
                if is_subset(r[1:], item):
                    m += 1
        return m/float(n)
 
    # ~~~~~~ Hilfsfunktion: ist die regel bereits bekannt?
    def is_known_rule(set, rule):
        for existing_rule in set.keys():
            if existing_rule[0] == rule[0]:
                if is_sameset(existing_rule[1:], rule[1:]):
                    return True

    # ~~~~~~ Hilfsfunktion: Gib eine Menge s am Bildschirm aus
    def dump(s, header, as_rule = False):
        temp = list(s)
        temp.sort()
        
        if as_rule:
            if not s:
                print "<h3>%s ist eine leere Menge!</h3><hr>" % (header)
                return
                
            print """<h3>%s Menge von %d Regeln:</h3>
<table width="800">
    <tr>
        <th width="100">support</th>
        <th width="100">confidence</th>
        <th width="100" align="right">Wenn</th>
        <th align="left">=&gt; Dann-Regel</th>
    </tr>
""" % (header, len(s))
            for r in temp:
                print """   <tr>
        <td align="right">%.0f%%</td>
        <td align="right">%.0f%%</td>
        <td align="right">%s</td>
        <td>=&gt; {%s}</td>
    </tr>""" % (support(r)*100, confidence(r)*100, r[0], ", ".join(r[1:]))
            print "</table>"
            
        else:
            print """<h3>%s Menge von %d Mengen:</h3>
<table width="800">
    <tr>
        <th width="100">support</th>
        <th align="left">Menge</th>
    </tr>
""" % (header, len(s))
            for r in temp:
                print """   <tr>
        <td align="right">%.0f%%</td>
        <td>{%s}</td>
    </tr>""" % (support(r)*100, ", ".join(r))
            print "</table>"

    # ~~~~~~ enlarge L[k]
    def enlarge_L(k):
        t = {}
        for item in C[k]:
            if support(item) >= min_support:
                t[item] = True
        L[k] = t.keys()
        dump(L[k], "L[%d] ist eine" % k, k>1)

    # ~~~~~~ select candidate C[k]
    def select_C(k):
        t = {}
        if k == 1:
            for s in D:
                for i in s:
                    t[(i,)] = True
                
        else:
            for x in L[1]:
                for tm in L[k-1]:
                    s = mergesets(x, tm)
                    if len(s) == k:
                        if not is_known_rule(t, s):
                            t[s] = True
        C[k] = t.keys()
        dump(C[k], "C[%d] ist eine" % k, k>1)

    dump(D, "Der input D ist eine")
    n, C, L = len(D), {}, {}
    for k in range(1, n):
        select_C(k)
        enlarge_L(k)
        if not L[k]:
            break

    # auswertung
    t = {}
    for i in range(2, k):
        for s in L[i]:
            t[s] = True
    L = t.keys()
    dump(L, "Die Ergebnismenge L ist eine", True)    

if __name__ == "__main__":

    # Eingabedaten
    apriori_algorithmus([('Brot', 'Butter', 'Wein', 'Kekse', 'Obst'),
    ('Brot', 'Zeitung', 'Wein', 'Milch', 'Zeitung'),
    ('Wein', 'Zucker', 'Mehl', 'Backpulver', 'Milch', 'Butter'),
    ('Wein', 'Cracker', 'Salz', 'Fisch', 'Fleisch', 'Butter', 'Milch', 'Schokolade'),
    ('Brot', 'Wein', 'Milch', 'Butter', 'Zeitung', 'Joghurt'),
    ('Zeitung', 'Bier', 'Wein', 'Brot'),
    ('Brot', 'Butter', 'Salz', 'Fleisch', 'Zucker', 'Mehl', 'Nsse', 'Backin'),
    ('Wein', 'Cracker', 'Brot', 'Fleisch', 'Schokolade', 'Zeitung')])


